import java.io.*;
import java.util.*;

/**
 * <p>Title: Scheduling Domain Problem Generator</p>
 *
 * <p>Description: </p>
 *
 * <p>Copyright: Ke Xu Copyright (c) 2006</p>
 *
 * <p>Company: Lehigh University</p>
 *
 * @author not attributable
 * @version 1.0
 */
public class HTNschprob {

    // number of problems to be generated
    public final static int PROB_NUM = 75;
    // number of tasks in each problem, TASK_NUM <= PART_NUM * 3
    public final static int TASK_NUM = 3;
    // number of parts in each problem, PART_NUM <= partList.length (26)
    public final static int PART_NUM = 2;
    public final static int TOTAL_PART = 20;
    // header of each problem name
    public final static String PROB_NAME = "schprob";

    // pool of parts
    private String[] partList = {"PART1", "PART2", "PART3", "PART4", "PART5",
                                 "PART6", "PART7", "PART8", "PART9", "PART10",
                                 "PART11", "PART12", "PART13", "PART14", "PART15",
                                 "PART16", "PART17", "PART18", "PART19", "PART20"};

    // pool of shapes for the init state
    private String[] shapeListS = {"CIRCULAR", "BOLONG"};
    // pool of colors for the init state
    private String[] colorListS = {"BLACK", "BLUE", "RED", "YELLOW"};
    // pool of surface-conditions for the init state
    private String[] surfaceListS = {"ROUGH", "SMOOTH", "POLISHED"};
    // pool of hole sizes for the init state
    private String[] sizeListS = {"ONE", "TWO", "THREE"};
    // pool of side of holes for the init state
    private String[] sideListS = {"FRONT", "BACK"};

    // pool of parts with conditions
    private Part[] partPool;

    // pool of temperatures for the init state
    private String[] lowTempListS = {"LowTemp1", "LowTemp2", "LowTemp3", "LowTemp4", "LowTemp5"};
    private String[] highTempListS = {"HighTemp1", "HighTemp2", "HighTemp3", "HighTemp4"}; // HighTemp5 will be "hot"
    private String[] genTempListS = {"Temp1", "Temp2", "Temp3", "Temp4", "Temp5",
                                     "Temp6", "Temp7", "Temp8", "Temp9", "Temp10"};

    // tempListS is the combination of lowTempListS, highTempListS and genTempListS.
    // If any of the above 3 lists is updated, tempListS also needs to be updated.
    private String[] tempListS = {"LowTemp1", "LowTemp2", "LowTemp3", "LowTemp4", "LowTemp5",
                                  "HighTemp1", "HighTemp2", "HighTemp3", "HighTemp4",
                                  "Temp1", "Temp2", "Temp3", "Temp4", "Temp5",
                                  "Temp6", "Temp7", "Temp8", "Temp9", "Temp10"};

     // pools of lowTempMachines
     private String[] polMachListS = {"polisher1", "polisher2", "polisher3"};
     private String[] punMachListS = {"punch1", "punch2", "punch3"};
     private String[] spMachListS = {"spray-painter1", "spray-painter2", "spray-painter3"};

     // pools of genTempMachines
     private String[] rolMachListS = {"roller1", "roller2", "roller3"};
     private String[] latMachListS = {"lathe1", "lathe2", "lathe3"};
     private String[] griMachListS = {"grinder1", "grinder2", "grinder3"};
     private String[] ipMachListS = {"immersion-painter1", "immersion-painter2", "immersion-painter3"};

    // Pools for generating tasks.
    // pool of shapes for tasks
    private String[] shapeListT = {"CYLINDRICAL"};

    // a 2-d array, which stores the init surface and color of each part
    private String[][] initList = new String[TOTAL_PART][2];

    // a public string
    public String publicTerm;

    // random seed
    private java.util.Random myRand = new java.util.Random( System.currentTimeMillis() );

    public HTNschprob() {
    }

    public void makeProb() {
        String probName = new String( "" ), outFile = new String( "" );
        int i, k;

        // Initialize parts
        partPool = new Part[TOTAL_PART];
        for( int cur_part = 0; cur_part < TOTAL_PART; cur_part++ )
        {
            partPool[cur_part] = new Part();
            partPool[cur_part].name = partList[cur_part];
            partPool[cur_part].shape = shapeListS[cur_part % shapeListS.length];
            partPool[cur_part].surface = surfaceListS[cur_part % surfaceListS.length];
            partPool[cur_part].paint = colorListS[cur_part % colorListS.length];
            partPool[cur_part].hole_size = sizeListS[cur_part % sizeListS.length];
            partPool[cur_part].hole_side = sideListS[cur_part % sideListS.length];
            partPool[cur_part].temp = tempListS[cur_part % tempListS.length];

            initList[cur_part][0] = partPool[cur_part].surface; // save init surface
            initList[cur_part][1] = partPool[cur_part].paint; // save init paint
        }

        try {
            for (k = 1; k <= PROB_NUM; k++)
            {
                probName = new String(PROB_NAME + String.valueOf( k ));
                outFile = new String("classes\\" + "schHTN\\" + probName + ".shp");
                BufferedWriter out = new BufferedWriter( new FileWriter( outFile ) );

                // head of the problem
                out.write("(defproblem " + probName + "\n");
                out.write("( \n");

                // output parts' info (shape, surface-condition, painted, has-hole, temperature)

                // randomly select parts from the part list
                int[] cur_parts = getParts();
                for ( i=0; i<PART_NUM; i++)
                {
                    out.write( partPool[cur_parts[i]].print() );
                }

                //output types of temperatures
                //generic temp
                for ( i=0; i<genTempListS.length; i++)
                {
                    out.write( "    (type " + genTempListS[i] + " Temp)\n");
                }
                //low temp
                for ( i=0; i<lowTempListS.length; i++)
                {
                    out.write( "    (type " + lowTempListS[i] + " LowTemp)\n");
                }
                //high temp
                for ( i=0; i<highTempListS.length; i++)
                {
                    out.write( "    (type " + highTempListS[i] + " HighTemp)\n");
                }

                // don't forget the "hot"
                out.write( "    (type hot HighTemp)\n");

                // output machines info (type, has-paint, can-orient)
                // no output for drill-press, it is not used in the domain
                String tmp;

                tmp = random( rolMachListS );
                out.write("    (type " + tmp + " Roller)\n");

                tmp = random( latMachListS );
                out.write("    (type " + tmp + " Lathe)\n");

                tmp = random( polMachListS );
                out.write("    (type " + tmp + " Polisher)\n");

                tmp = random( griMachListS );
                out.write("    (type " + tmp + " Grinder)\n");

                tmp = random( punMachListS );
                out.write("    (type " + tmp + " Punch)\n");
                out.write("    (CAN-ORIENT " + tmp + " BACK)\n");
                out.write("    (CAN-ORIENT " + tmp + " FRONT)\n");
                out.write("    (HAS-BIT " + tmp + " THREE)\n");
                out.write("    (HAS-BIT " + tmp + " TWO)\n");
                out.write("    (HAS-BIT " + tmp + " ONE)\n");

                tmp = random( spMachListS );
                out.write("    (type " + tmp + " Spray-Painter)\n");
                out.write("    (HAS-PAINT " + tmp + " YELLOW)\n");
                out.write("    (HAS-PAINT " + tmp + " BLUE)\n");
                out.write("    (HAS-PAINT " + tmp + " BLACK)\n");
                out.write("    (HAS-PAINT " + tmp + " RED)\n");

                tmp = random( ipMachListS );
                out.write("    (type " + tmp + " Immersion-Painter)\n");
                out.write("    (HAS-PAINT " + tmp + " YELLOW)\n");
                out.write("    (HAS-PAINT " + tmp + " BLUE)\n");
                out.write("    (HAS-PAINT " + tmp + " BLACK)\n");
                out.write("    (HAS-PAINT " + tmp + " RED)\n");

                out.write(") \n");
                out.write("((achieve-goals (  list \n");

                // output tasks
                String[] tasks = getTasks( cur_parts );
                for ( i = 0; i <tasks.length; i++ )
                {
                    out.write( tasks[i] );
                }

                out.write(")))) \n");
                out.close();
            }
        } catch (IOException e) {
            System.out.println( "Error when writing: " + outFile + " -- " + e );
        }
    }

    public int[] getParts()
    {
        ArrayList list = new ArrayList();
        int[] parts = new int[PART_NUM];
        int i = 0;

        while ( list.size() < PART_NUM )
        {
            int tmp = random( TOTAL_PART );
            String number = Integer.toString(tmp);
            if ( !list.contains( number ))
            {
                parts[i] = tmp;
                i++;
                list.add(number);
            }
        }

        return parts;
    }

    public String[] getTasks( int[] pl )
    { // randomly select tasks from the task pool
        String[] taskPool = makeTaskPool( pl );
        ArrayList list = new ArrayList();
        String[] tasks = new String[TASK_NUM];
        int i;

        while ( list.size() < TASK_NUM )
        {
            String tmp = random( taskPool );
            if ( !list.contains( tmp ) )
                list.add( tmp );
        }

        for ( i=0; i<TASK_NUM; i++ )
        {
            tasks[i] = (String)list.get( i );
        }

        return tasks;
    }

    public String[] makeTaskPool( int[] pl )
    { // make the task pool: 3 different tasks for each part.
      // a task should not exist in the init state (to avoid "do-nothing" plan)

        String[] taskPool = new String[PART_NUM*3];
        for ( int i=0; i<PART_NUM; i++  )
        {
            String[] tmp = partTasks( partList[pl[i]], pl[i] );
            taskPool[i*3] = tmp[0];
            taskPool[i*3+1] = tmp[1];
            taskPool[i*3+2] = tmp[2];
        }
        return taskPool;
    }

    public String[] partTasks( String part, int partIndex )
    { // generate 3 tasks for a part
        int k, index;
        String[] surfaceListT = new String[surfaceListS.length - 1];
        String[] colorListT = new String[colorListS.length - 1];

        // make the surface pool for the tasks
        index = 0;
        for ( k=0; k<surfaceListS.length; k++ )
        {
            // if this surface is not used in init state, use it for task
            if ( !surfaceListS[k].equalsIgnoreCase( initList[partIndex][0] ) )
            {
                surfaceListT[index] = surfaceListS[k];
                index++;
            }
        }

        // make the color pool for the tasks
        index = 0;
        for ( k=0; k<colorListS.length; k++ )
        {
            // if this color is not used in init state, use it for task
            if ( !colorListS[k].equalsIgnoreCase( initList[partIndex][1] ) )
            {
                colorListT[index] = colorListS[k];
                index++;
            }
        }

        //generate 3 tasks for the part
        String[] pt = new String[3];
        pt[0] = partShape( part, shapeListT );
        pt[1] = partColor( part, colorListT );
        pt[2] = partSurface( part, surfaceListT );
        return pt;
    }

    public String partShape( String part, String[] pool )
    { // generate a shape atom of a part, using a shape pool
        String randomTerm = random( pool );
        String line = new String( "    (SHAPE " + part + " " + randomTerm + ") \n" );
        return line;
    }

    public String partSurface(String part, String[] pool )
    { // generate a surface atom of a part, using a surface pool
        publicTerm = random( pool );
        String line = new String("    (SURFACE-CONDITION " + part + " " + publicTerm + ") \n");
        return line;
    }

    public String partColor( String part, String[] pool )
    { // generate a color atom of a part, using a color pool
        publicTerm = random( pool );
        String line = new String("    (PAINTED " + part + " " + publicTerm + ") \n");
        return line;
    }

    public String partHole( String part, String[] sizePool, String[] sidePool )
    { // generate a hole atom of a part
        String randomTerm1 = random( sizePool );
        String randomTerm2 = random( sidePool );
        String line = new String("    (HAS-HOLE " + part + " " + randomTerm1 + " " + randomTerm2 +  ") \n");
        return line;
    }

    public String partTemp( String part, String[] pool)
    { // generate a temperature atom of a part
        String randomTerm = random( pool );
        String line = new String("    (TEMPERATURE " + part + " " + randomTerm + ") \n");
        return line;
    }

     public String random( String[] list )
    {
        int index = random( list.length );
        String result = list[index];
        return result;
    }

    public int random( int upperbound )
    { // returns an int in the range from [0,upperbound)
        return (myRand.nextInt(upperbound));
    }

    public static void main( String[] args )
    {
        HTNschprob probMaker = new HTNschprob();
        probMaker.makeProb();
    }

    class Part
    {
        String name, shape, surface, paint, hole_size, hole_side, temp;
        /*
             (SHAPE R CIRCULAR)
             (SURFACE-CONDITION R ROUGH)
             (PAINTED R YELLOW)
             (HAS-HOLE R TWO BACK)
             (TEMPERATURE R Temp5)
         */

        public String print()
        {
            String partInfo = new String(
                     "    (SHAPE " + name + " " + shape + ")\n"
                   + "    (SURFACE-CONDITION " + name + " " + surface + ")\n"
                   + "    (PAINTED " + name + " " + paint + ")\n"
                   + "    (HAS-HOLE " + name + " " + hole_size + " " + hole_side + ")\n"
                   + "    (TEMPERATURE " + name + " " + temp + ")\n");
            return partInfo;
        }
    }

}
